#ifndef SERVERLIB
  #define SERVERLIB

#include <ESP8266WiFi.h>

const int ERROR_NUMBER = -9999;

enum OBJECTTYPE {
    OBJECTTYPE_KEYVALUE_STRING,
    OBJECTTYPE_KEYVALUE_INT,
    OBJECTTYPE_PINS_ARRAY,
    OBJECTTYPE_PIN,
    OBJECTTYPE_PINS,
    OBJECTTYPE_DATE,
    OBJECTTYPE_WHITELIST_ARRAY,
    OBJECTTYPE_STATUS,
    OBJECTTYPE_LOG_PINS,
    OBJECTTYPE_INFORMATION,
    OBJECTTYPE_WARNING,
    OBJECTTYPE_ERROR,
    /*add next type above this line*/
    OBJECTTYPE_COUNT
};

enum JSONTYPE {
    KEYVALUE_STRING,
    KEYVALUE_INT,
    KEYVALUE_DOUBLE
};

/// <summary>
/// Possible types of a pin
/// </summary>
enum PINTYPE {
    PINTYPE_INPUT_VIRTUAL=-1,
    PINTYPE_INPUT_ANALOG,   /*Read method analogRead shall be used*/
    PINTYPE_INPUT_DIGITAL,  /*Read method digitalRead shall be used*/
    PINTYPE_OUTPUT_ANALOG,  /*Write method analogWrite shall be used*/
    PINTYPE_OUTPUT_DIGITAL /*Write method digitalWrite shall be used*/
};



#ifndef CODE_BLOCK_LinkedList
template<class T>
struct ListNode
{
    T data;
    ListNode<T> *next;
};

template <typename T>
class LinkedList {

protected:
    int _size;
    ListNode<T> *root;
    ListNode<T> *last;

    // Helps "get" method, by saving last position
    ListNode<T> *lastNodeGot;
    int lastIndexGot;
    // isCached should be set to FALSE
    // everytime the list suffer changes
    bool isCached;

    ListNode<T>* getNode(int index);

public:
    LinkedList();
    ~LinkedList();

    /*
    Returns current size of LinkedList
    */
    virtual int size();
    /*
    Adds a T object in the specified index;
    Unlink and link the LinkedList correcly;
    Increment _size
    */
    virtual bool add(int index, T);
    /*
    Adds a T object in the end of the LinkedList;
    Increment _size;
    */
    virtual bool add(T);
    /*
    Adds a T object in the start of the LinkedList;
    Increment _size;
    */
    virtual bool unshift(T);
    /*
    Set the object at index, with T;
    Increment _size;
    */
    virtual bool set(int index, T);
    /*
    Remove object at index;
    If index is not reachable, returns false;
    else, decrement _size
    */
    virtual T remove(int index);
    /*
    Remove last object;
    */
    virtual T pop();
    /*
    Remove first object;
    */
    virtual T shift();
    /*
    Get the index'th element on the list;
    Return Element if accessible,
    else, return false;
    */
    virtual T get(int index);

    /*
    Clear the entire array
    */
    virtual void clear();

};

// Initialize LinkedList with false values
template<typename T>
LinkedList<T>::LinkedList()
{
    root = NULL;
    last = NULL;
    _size = 0;

    lastNodeGot = root;
    lastIndexGot = 0;
    isCached = false;
}

// Clear Nodes and free Memory
template<typename T>
LinkedList<T>::~LinkedList()
{
    ListNode<T>* tmp;
    while (root != NULL)
    {
        tmp = root;
        root = root->next;
        delete tmp;
    }
    last = NULL;
    _size = 0;
    isCached = false;
}

/*
Actualy "logic" coding
*/

template<typename T>
ListNode<T>* LinkedList<T>::getNode(int index) {

    int _pos = 0;
    ListNode<T>* current = root;

    // Check if the node trying to get is
    // immediatly AFTER the previous got one
    if (isCached && lastIndexGot <= index) {
        _pos = lastIndexGot;
        current = lastNodeGot;
    }

    while (_pos < index && current) {
        current = current->next;

        _pos++;
    }

    // Check if the object index got is the same as the required
    if (_pos == index) {
        isCached = true;
        lastIndexGot = index;
        lastNodeGot = current;

        return current;
    }

    return false;
}

template<typename T>
int LinkedList<T>::size() {
    return _size;
}

template<typename T>
bool LinkedList<T>::add(int index, T _t) {

    if (index >= _size)
        return add(_t);

    if (index == 0)
        return unshift(_t);

    ListNode<T> *tmp = new ListNode<T>(),
        *_prev = getNode(index - 1);
    tmp->data = _t;
    tmp->next = _prev->next;
    _prev->next = tmp;

    _size++;
    isCached = false;

    return true;
}

template<typename T>
bool LinkedList<T>::add(T _t) {

    ListNode<T> *tmp = new ListNode<T>();
    tmp->data = _t;
    tmp->next = NULL;

    if (root) {
        // Already have elements inserted
        last->next = tmp;
        last = tmp;
    }
    else {
        // First element being inserted
        root = tmp;
        last = tmp;
    }

    _size++;
    isCached = false;

    return true;
}

template<typename T>
bool LinkedList<T>::unshift(T _t) {

    if (_size == 0)
        return add(_t);

    ListNode<T> *tmp = new ListNode<T>();
    tmp->next = root;
    tmp->data = _t;
    root = tmp;

    _size++;
    isCached = false;

    return true;
}

template<typename T>
bool LinkedList<T>::set(int index, T _t) {
    // Check if index position is in bounds
    if (index < 0 || index >= _size)
        return false;

    getNode(index)->data = _t;
    return true;
}

template<typename T>
T LinkedList<T>::pop() {
    if (_size <= 0)
        return T();

    isCached = false;

    if (_size >= 2) {
        ListNode<T> *tmp = getNode(_size - 2);
        T ret = tmp->next->data;
        delete(tmp->next);
        tmp->next = NULL;
        last = tmp;
        _size--;
        return ret;
    }
    else {
        // Only one element left on the list
        T ret = root->data;
        delete(root);
        root = NULL;
        last = NULL;
        _size = 0;
        return ret;
    }
}

template<typename T>
T LinkedList<T>::shift() {
    if (_size <= 0)
        return T();

    if (_size > 1) {
        ListNode<T> *_next = root->next;
        T ret = root->data;
        delete(root);
        root = _next;
        _size--;
        isCached = false;

        return ret;
    }
    else {
        // Only one left, then pop()
        return pop();
    }

}

template<typename T>
T LinkedList<T>::remove(int index) {
    if (index < 0 || index >= _size)
    {
        return T();
    }

    if (index == 0)
        return shift();

    if (index == _size - 1)
    {
        return pop();
    }

    ListNode<T> *tmp = getNode(index - 1);
    ListNode<T> *toDelete = tmp->next;
    T ret = toDelete->data;
    tmp->next = tmp->next->next;
    delete(toDelete);
    _size--;
    isCached = false;
    return ret;
}

template<typename T>
T LinkedList<T>::get(int index) {
    ListNode<T> *tmp = getNode(index);

    return (tmp ? tmp->data : T());
}

template<typename T>
void LinkedList<T>::clear() {
    while (size() > 0)
        shift();
}
#endif //CODE_BLOCK_LinkedList

struct PinValue
{
    int pinNumber;
    int pinValue;

};

class KeyValue {
private:
    unsigned int mKeyValueType;
    String mKey;
    String mstrValue;
    int    miValue;
    double  mdValue;
    void set(unsigned int keyValueType, String key, String strValue, int iValue, double dValue);
public:

    KeyValue(String key, String value);
    KeyValue(String key, int value);
    KeyValue(String key, double value);
    String getKey();
    String getValueString(boolean addDoubleQuotationIfString);
    //returns a keyvalue string pair.
    //on error "" is returned.
    String toJson();
};

class Json : public LinkedList<KeyValue*> {

private:
public:
    Json();
    ~Json();

    bool add(String key, String value) { return LinkedList<KeyValue*>::add(new KeyValue(key, value)); }
    bool add(String key, int value) { return LinkedList<KeyValue*>::add(new KeyValue(key, value)); }
    bool add(String key, double value) { return LinkedList<KeyValue*>::add(new KeyValue(key, value)); }
    String toJson();
    //deletes all objects and destoys the list;
    void destory();
};

class GTime {
    ///////////////////////////////////////////////////////////
private:
    int mYear = 0;
    int mMonth = 0;
    int mDay = 0;
    int mHours = 0;
    int mMinutes = 0;
    int mSeconds = 0;
public:
    GTime() { }
    /// <summary>
    /// Sets time values from a given string
    /// </summary>
    /// <param name="strTime">
    /// A string with date and time.
    /// The string needs to be formatted like this: 
    /// Fri, 15 Jul 2016 11:08:12 GMT
    /// </param>
    /// <returns></returns>
    boolean setTime(String strTime);
    /// <summary>
    /// Converts a 3 letter english month name to the number of the month in the year.
    /// </summary>
    /// <param name="month">
    /// A string of length 3, which represents the month.  
    /// For example "JAN" for january.</param>
    /// <returns>
    /// success : the number of the month, where 1 is january.
    /// fail    : -1 
    /// </returns>  
    int strToMonth(String month);
    /// <summary>
    /// Converts a String to a number positive number.
    /// Negative numbers are considered invalid.
    /// </summary>
    /// <param name="str">The string to be converted to a number.</param>
    /// <returns>
    /// Success: The converted number.
    /// Fail   : -9999
    /// </returns>
    static int toNumber(String str);
    /// <summary>
    /// Returns date and time values as a english date time string.
    /// </summary>
    /// <returns>Date and time string on the format "MM/DD/YY hh:mm:ss"</returns>
    String toString();
    /// <summary>
    /// Returns date and time values as a icelandic date time string.
    /// </summary>
    /// <returns>Date and time string on the format "DD.MM.YY hh:mm:ss"</returns>
    String toStreng();
    /// <summary>
    /// Returns the date and time values as an JSON object.
    /// </summary>
    /// <returns>A String formatted as an JSON object</returns>
    String toJson();
};

class IPAddressList : public LinkedList<IPAddress*> {
private:
    /// <summary>
    /// //The cleanup function used by the list's deconstructor;
    /// </summary>
    void destory();

public:
    /// <summary>
    /// Add an ip address by providing four numbers of the IP addres each in the range of 0 - 255
    /// </summary>
    /// <param name="first_octet">First number of the ip address </param>
    /// <param name="second_octet">Second number of the ip address</param>
    /// <param name="third_octet">Third number of the ip address </param>
    /// <param name="fourth_octet">Fourth number of the ip address</param>
    /// <returns></returns>
    bool add(uint8_t first_octet, uint8_t second_octet, uint8_t third_octet, uint8_t fourth_octet) { return add(IPAddress(first_octet, second_octet, third_octet, fourth_octet)); };
    /// <summary>
    /// Add an ip address to the list by providing a IP address
    /// </summary>
    /// <param name="address">The IPAddress object to be added to the list</param>
    /// <returns>True if the add succeded, otherwise false</returns>
    bool add(IPAddress address);
    /// <summary>
    /// Add an ip address to the list by providing a string with a valid IP address
    /// </summary>
    /// <param name="strIpAddress">
    ///     a string with a valid IP address.  
    ///     The ip address "0.0.0.0" will be considered as an invalid ipaddress
    /// </param>
    /// <returns>True if the add succeded, otherwise false</returns>
    bool add(const char* strIpAddress);
    /// <summary>
    /// Checks if a ip address exists in the list
    /// </summary>
    /// <param name="address">The ip address to search for</param>
    /// <returns>True if the add ip address was found in the list, otherwise false.</returns>
    bool exists(IPAddress address);
    bool exists(String strIpaddress);

    /// <summary>
    /// Searches for the index of a ip address in the list.
    /// </summary>
    /// <param name="address">The ip address to search for.</param>
    /// <returns>
    /// Success:The index of the ip address in the list.  
    /// fail   : -1 if string is not found.</returns>
    int indexOf(IPAddress address);
    bool isEmpty();
    bool remove(const char *strIpAddress);
    bool remove(IPAddress address);

    //returns the list as a jsonObject array

    /// <summary>
    /// Bundles all Ip addresses into a JSON array
    /// </summary>
    /// <returns>
    /// A string containing a valid JSON array of ip addresses.
    /// Example of a returned string: ["192.168.1.54","10.1.1.15","10.1.1.1","255.255.255.0"]</returns>
    String toJson();

    /// <summary>
    /// The deconstructor, which cleans up when the list is no longer needed.
    /// </summary>
    ~IPAddressList();
};



















class GPin {
private:
    int mNumber;
    int mValue;
    PINTYPE mType;
    char *mName;
#ifdef ESP32 
    //#define LEDC_TIMER_13_BIT  13
    //#define LEDC_BASE_FREQ     8100
    #define LEDC_TIMER_13_BIT  13
    #define LEDC_BASE_FREQ     8100
    int mChannel;
    void ledcAnalogWrite(uint8_t channel, uint32_t value, uint32_t valueMax = 255);
    void analogWriteEsp32();
#endif
    void init(const char*strPinName, PINTYPE pinType, int pinNumber, int pinValue);
    void set(int number, int value);
    String jsonKeyValue(String key, int value);
    void destroy();
public:
#ifdef ESP32 
    GPin(const char*strPinName, PINTYPE pinType, int pinNumber, int pinValue, uint8_t pinChannel);
#else
    GPin(const char *strPinName, PINTYPE pinType, int pinNumber, int pinValue);
#endif    
    ~GPin();
    
    void setName(const char*strPinName);
    String getName();
    void setValue(int value);
    int getValue(bool readValueFromHardware = true);
    int getNumber();
    int getType();
    String toJson();
    String toJsonPinout();
};

/// <summary>
/// This class stores all pins
/// </summary>
class GPins {

private:
    int mLength = 0;
#ifdef ESP32 
    int mChannelCount = 0;
#endif
    GPin *mPins[30];//todo: I make this dynamic, instead of a fixed size

public:
    GPins() { mLength = 0; }
    //todo: I' don't need a deconstructur but I should make one
#ifdef ESP32 
    int addPin(const char *strPinName, PINTYPE pinType, int pinNumber, int pinValue);
#else
    int addPin(const char *strName, PINTYPE pinType, int pinNumber, int pinValue);
#endif
    //sets the value of a pin with the given number.
    boolean setValue(int pinNumber, int newValue);
    //searches for a pin of a specific number and returns it's index in the array
    int indexOf(int pinNumber);
    //returns true if a pin with the specific number exists in the array
    boolean exists(int pinNumber);
    //returns NULL if pin is not found at a given index
    GPin *get(int pinNumber);

	//get pin by index, be careful when using this function
	GPin *getByIndex(int index){if (index<0) return NULL;  return this->mPins[index];}
    //returns the value of the given pin.  Returns -1 if pinNumber was not found
    int getValue(int pinNumber);
    int count() { return mLength; }

    /// <summary>
    /// Returns all pin values in a json array
    /// a key-value Json object with the '{' and '}' around it.
    /// where first key is the first in the index with the key as the GPO key
    /// and the value is the last value set to that key.
    /// </summary>
    /// <returns></returns>
    String toJson();

    /// <summary>
    /// Creates a JSON object containg all pins name and their number.
    /// </summary>
    /// <returns>A string formatted as a JSON object which contains all pin names and number. </returns>
    String JsonPinout();
};

class GUrl {
private:
    int mLength = 0;
public:
    GUrl() { }
    String getAfter(String str, String afterMe);
    int toNumber(String str);
    boolean extractAndSetPinsAndValues(const char *unParsedJson, GPins *pinnar);
    uint8_t extractPinValues(const char *unParsedJson, PinValue pinValuesArray[], uint8_t arrayLength);
    String removeLastSpaceIfExists(String str);
    String jsonKeyValue(String key, String value);
    /// <summary>
    /// Creates a string with a key and a value, which can be used when populating a JSON object.
    /// </summary>
    /// <param name="name">Name of the key</param>
    /// <param name="value">The integer value</param>
    /// <returns>
    /// A string with a key and a value.  
    /// For example: 
    /// "args":12
    /// </returns>
    String jsonKeyValue(String key, int value);
    String jsonObjectType(unsigned int uiType);
    String makeStatusResponceJson(String jsonPins, String jsonWhitelist, String jsonDate);
    String makePostLogPinsJson(String deviceId, String jsonPins);
    /// <summary>
    /// Formats a http status code
    /// </summary>
    /// <param name="uiStatusCode">Number of the http status code to format</param>
    /// <returns>A string with the http statuscode number and the status text.</returns>
    String makeHttpStatusCodeString(unsigned int uiStatusCode);
    String extractAndReturnIPaddress(const char *unParsedJson);
    String jsonRoot(unsigned int uiType, String key, String value);
};

#endif
